from attr import dataclass
from ..internal.symbol_export_sub_chunk import (
    new_sub_chunk as nsc,
    release_sub_chunk,
    sub_chunk_block,
    sub_chunk_empty,
    sub_chunk_equals,
    sub_chunk_set_block,
)


class SubChunkBase:
    """SubChunkBase is the base implement of a Minecraft sub chunk."""

    _sub_chunk_id: int

    def __init__(self):
        self._sub_chunk_id = -1

    def __del__(self):
        if self._sub_chunk_id >= 0 and not release_sub_chunk is None:
            release_sub_chunk(self._sub_chunk_id)


class SubChunk(SubChunkBase):
    """
    SubChunk is a cube of blocks located in a chunk.
    It has a size of 16x16x16 blocks and forms part of a stack that forms a Chunk.
    """

    def __init__(self):
        super().__init__()

    def equals(self, another_sub_chunk: SubChunkBase) -> tuple[bool, bool]:
        """Equals returns if the sub chunk passed is equal to the current one.

        Args:
            another_sub_chunk (SubChunkBase): The sub chunk passed.

        Returns:
            tuple[bool, bool]: The first element refer to the compare result,
                               and the second one refer to if their have any
                               any error occurred.
                               If the second one is False, then it means current
                               sub chunk or another_sub_chunk is not found.
        """
        result = sub_chunk_equals(self._sub_chunk_id, another_sub_chunk._sub_chunk_id)
        return (result == 1, result != -1)

    def empty(self) -> tuple[bool, bool]:
        """
        empty checks if the SubChunk is considered empty.
        This is the case if the SubChunk has 0 block storages or if it has a single one that is completely filled with air.

        Returns:
            tuple[bool, bool]: If current sub chunk is not found, return (False, False).
                               Otherwise, return the result (is empty or not) and True.
        """
        result = sub_chunk_empty(self._sub_chunk_id)
        return (result == 1, result != -1)

    def block(self, x: int, y: int, z: int, layer: int) -> int:
        """
        block returns the runtime ID of the block located at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.
            layer (int): The layer that the target block is in.

        Returns:
            int: Return the block runtime ID of target block.
                 If current sub chunk is not found, then return -1.
                 Note that if no sub chunk exists at the given y, the block is assumed to be air.
        """
        return sub_chunk_block(self._sub_chunk_id, x, y, z, layer)

    def set_block(self, x: int, y: int, z: int, layer: int, block_runtime_id: int):
        """
        set_block sets the given block runtime ID at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.
            layer (int): The layer that the target block is in.
            block_runtime_id (int): The block runtime id of target block will be.

        Raises:
            Exception: When failed to set block.
        """
        err = sub_chunk_set_block(self._sub_chunk_id, x, y, z, layer, block_runtime_id)
        if len(err) > 0:
            raise Exception(err)


@dataclass
class SubChunkWithIndex:
    """SubChunkWithIndex represents a sub chunk and its index in a whole chunk where this sub chunk is in.

    Args:
        sub_chunk (SubChunk): The sub chunk.
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0
    """

    index: int = 0
    sub_chunk: SubChunk = SubChunk()


def new_sub_chunk() -> SubChunk:
    """
    NewSubChunk creates a new sub chunk.
    All sub chunks should be created through this function.

    Returns:
        SubChunk: A new sub chunk that full of air.
    """
    s = SubChunk()
    s._sub_chunk_id = nsc()
    return s



from attr import dataclass
import nbtlib


DIMENSION_OVERWORLD = 0
DIMENSION_NETHER = 1
DIMENSION_END = 2


@dataclass
class ChunkPos:
    """
    ChunkPos holds the position of a chunk. The type is provided as a utility struct for keeping track of a
    chunk's position. Chunks do not themselves keep track of that. Chunk positions are different from block
    positions in the way that increasing the X/Z by one means increasing the absolute value on the X/Z axis in
    terms of blocks by 16.
    """

    x: int = 0
    z: int = 0


@dataclass
class SubChunkPos:
    """
    SubChunkPos holds the position of a sub-chunk. The type is provided as a utility struct for keeping track of a
    sub-chunk's position. Sub-chunks do not themselves keep track of that. Sub-chunk positions are different from
    block positions in the way that increasing the X/Y/Z by one means increasing the absolute value on the X/Y/Z axis in
    terms of blocks by 16.
    """

    x: int = 0
    y: int = 0
    z: int = 0


@dataclass
class Range:
    """
    Range represents the height range of a Dimension in blocks. The first value
    of the Range holds the minimum Y value, the second value holds the maximum Y
    value.
    """

    start_range: int = 0
    end_range: int = 0


class Dimension:
    """
    Dimension is a dimension of a World. It influences a variety of
    properties of a World such as the building range, the sky colour and the
    behaviour of liquid blocks.
    """

    dm: int = 0

    def __init__(self, dm: int):
        """Init a new dimension represent.

        Args:
            dm (int): The id of this dimension.
        """
        self.dm = dm

    def range(self) -> Range:
        """range return the range that player could build block in this dimension.

        Returns:
            Range: The range that player could build block in this dimension.
                   If this dimension is not standard dimension, then redirect
                   to overworld range.
        """
        match self.dm:
            case 0:
                return Range(-64, 319)
            case 1:
                return Range(0, 127)
            case 2:
                return Range(0, 255)
            case _:
                return Range(-64, 319)

    def height(self) -> int:
        """
        height returns the height of this dimension.
        For example, the height of overworld is 384
        due to "384 = 319 - (-64) + 1", and 319 is
        the max Y that overworld could build, and -64
        is the min Y that overworld could build.

        Returns:
            int: The height of this dimension.
                 If this dimension is not standard dimension, then redirect
                 to overworld height.
        """
        match self.dm:
            case 0:
                return 384
            case 1:
                return 128
            case 2:
                return 256
            case _:
                return 384

    def __str__(self) -> str:
        match self.dm:
            case 0:
                return "Overworld"
            case 1:
                return "Nether"
            case 2:
                return "End"
            case _:
                return f"Custom (id={self.dm})"


@dataclass
class BlockStates:
    """BlockState holds a combination of a name and properties."""

    Name: str = ""
    States: nbtlib.tag.Compound = nbtlib.tag.Compound()


@dataclass
class HashWithPosY:
    Hash: int = 0
    PosY: int = 0



import nbtlib
from .define import BlockStates, Range
from ..world.sub_chunk import SubChunk, SubChunkWithIndex
from ..internal.symbol_export_conversion import (
    runtime_id_to_state as rits,
    state_to_runtime_id as stri,
    sub_chunk_network_payload as scnp,
    from_sub_chunk_network_payload as fscnp,
    sub_chunk_disk_payload as scdp,
    from_sub_chunk_disk_payload as fscdp,
)


def runtime_id_to_state(
    block_runtime_id: int,
) -> tuple[BlockStates | None, bool]:
    """runtime_id_to_state convert block runtime id to a BlockStates.

    Args:
        block_runtime_id (int): The runtime id of target block.

    Returns:
        tuple[BlockStates | None, bool]: If not found, return (None, False).
                                         Otherwise, return BlockStates and True.
    """
    name, states, success = rits(block_runtime_id)
    if not success:
        return (None, False)
    return (BlockStates(name, states), True)  # type: ignore


def state_to_runtime_id(
    block_name: str, block_states: nbtlib.tag.Compound
) -> tuple[int, bool]:
    """
    state_to_runtime_id convert a block which name is block_name
    and states is block_states to its block runtime id represent.

    Args:
        block_name (str): The name of this block.
        block_states (nbtlib.tag.Compound): The block states of this block.

    Returns:
        tuple[int, bool]: If not found, return (0, False).
                          Otherwise, return its block runtime id and True.
    """
    block_runtime_id, success = stri(block_name, block_states)
    if not success:
        return (0, False)
    return (block_runtime_id, True)


def sub_chunk_network_payload(sub_chunk: SubChunk, r: Range, index: int) -> bytes:
    """
    sub_chunk_network_payload encodes sub_chunk to its payload represent that could use on network sending.

    Args:
        sub_chunk (SubChunk): The sub chunk want to encode.
        r (Range): The whole chunk range where this sub chunk is in.
                   For overworld, it is Range(-64, 319).
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0


    Returns:
        bytes: The bytes represent of this sub chunk, and could especially sending on network.
               Therefore, this is a Network encoding sub chunk payload.
    """
    return scnp(sub_chunk._sub_chunk_id, r.start_range, r.end_range, index)


def from_sub_chunk_network_payload(
    r: Range, payload: bytes
) -> tuple[SubChunkWithIndex | None, bool]:
    """from_sub_chunk_network_payload decode a Network encoding sub chunk and return its python represent.

    Args:
        r (Range): The whole chunk range where this sub chunk is in.
                   For overworld, it is Range(-64, 319).
        payload (bytes): The bytes of this sub chunk, who with a Network encoding.

    Returns:
        tuple[SubChunkWithIndex | None, bool]: If failed to decode, then return (None, False).
                                               Otherwise, return decoded sub chunk and True.
    """
    index, sub_chunk_id, success = fscnp(r.start_range, r.end_range, payload)
    if not success:
        return (None, False)
    s = SubChunk()
    s._sub_chunk_id = sub_chunk_id
    return (SubChunkWithIndex(index, s), True)


def sub_chunk_disk_payload(sub_chunk: SubChunk, r: Range, index: int) -> bytes:
    """
    sub_chunk_disk_payload encodes sub_chunk to its payload represent under Disk encoding.
    That means the returned bytes could save to disk if its len bigger than 0.

    Args:
        sub_chunk (SubChunk): The sub chunk want to encode.
        r (Range): The whole chunk range where this sub chunk is in.
                   For overworld, it is Range(-64, 319).
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0


    Returns:
        bytes: The bytes represent of this sub chunk, who with a Disk encoding.
    """
    return scdp(sub_chunk._sub_chunk_id, r.start_range, r.end_range, index)


def from_sub_chunk_disk_payload(
    r: Range, payload: bytes
) -> tuple[SubChunkWithIndex | None, bool]:
    """from_sub_chunk_disk_payload decode a Disk encoding sub chunk and return its python represent.

    Args:
        r (Range): The whole chunk range where this sub chunk is in.
                   For overworld, it is Range(-64, 319).
        payload (bytes): The bytes of this sub chunk, who with a Disk encoding.

    Returns:
        tuple[SubChunkWithIndex | None, bool]: If failed to decode, then return (None, False).
                                               Otherwise, return decoded sub chunk and True.
    """
    index, sub_chunk_id, success = fscdp(r.start_range, r.end_range, payload)
    if not success:
        return (None, False)
    s = SubChunk()
    s._sub_chunk_id = sub_chunk_id
    return (SubChunkWithIndex(index, s), True)
